"""
TSSV: Targeted characterisation of short structural variation.


The library file consists of four tab-separated columns:
- name of the marker pair
- marker 1
- marker 2
- space-separated description of the expected pattern

If the -d option is used, a folder will be created containing the library
table and per marker a subfolder containing the new alleles and split FASTA
files.

Copyright (c) 2016 Leiden University Medical Center <humgen@lumc.nl>
Copyright (c) 2016 Jeroen F.J. Laros <j.f.j.laros@lumc.nl>
Copyright (c) 2016 Jerry Hoogenboom <j.hoogenboom@nfi.minvenj.nl>
Copyright (c) 2012 Jaap W.F. van der Heijden

Licensed under the MIT license, see the LICENSE file.
"""

import argparse
import os

__version_info__ = ('0', '4', '0')

__version__ = '.'.join(__version_info__)
__author__ = 'LUMC, Jeroen F.J. Laros'
__contact__ = 'J.F.J.Laros@lumc.nl'
__homepage__ = 'https://git.lumc.nl/j.f.j.laros/tssv'

usage = __doc__.split("\n\n\n")

class ProtectedFileType(argparse.FileType):
    def __call__(self, string):
        if 'w' in self._mode and os.path.exists(string):
            raise IOError('failed to create "{}": file exists.'.format(string))
        return super(ProtectedFileType, self).__call__(string)

def doc_split(func):
    return func.__doc__.split("\n\n")[0]

def version(name):
    return "%s version %s\n\nAuthor   : %s <%s>\nHomepage : %s" % (name,
        __version__, __author__, __contact__, __homepage__)



from . import sg_align


def align_pair(reference, reference_rc, pair, indel_score=1):
    """
    Align a pair of markers to the forward reference sequence. The reverse
    complement is used to align the second element of the pair (which is also
    reverse complemented).

    :arg reference: Reference sequence to align to.
    :type reference: str
    :arg reference_rc: Reverse complement of the reference sequence.
    :type reference_rc: str
    :arg pair: A pair (forward, reverse) of markers to align.
    :type pair: list[str, str]
    :arg indel_score: Penalty score for insertions and deletions per
    nucleotide (optional, default: 1)
    :type indel_score: int

    :returns: A tuple (score, position) of the best alignment.
    :rtype: tuple(int, int)
    """
    right_alignment = sg_align.align(reference_rc, pair[1], indel_score)
    left_alignment = sg_align.align(reference, pair[0], indel_score)

    return ((left_alignment.distance, left_alignment.position),
        (right_alignment.distance, len(reference) - right_alignment.position))



#!/usr/bin/env python

"""
Convert a csv files containing alleles and counts to HGVS descriptions of
alleles and single variants. Also report statistics about the variant types.


The input file is typically one of the output files of tssv.
"""

import argparse
import collections
import numpy
import sys

from suds import client

URL = "https://mutalyzer.nl/services/?wsdl"

def write_table(data, title, report_handle, minimum):
    """
    Write a table to a file.

    :arg data: Dictionary containing counts per type.
    :type data: dict
    :arg title: Name of the first column.
    :type title: str
    :arg report_handle: Open writeable handle to the report file.
    :type report_handle: stream
    :arg minimum: Minimum count.
    :type minimum: int
    """
    report_handle.write("%s\ttotal\tforward\treverse\n" % title)

    for i in sorted(data, key=lambda x: x[0], reverse=True):
        if data[i][0] < minimum:
            return

        report_handle.write("%s\t%i\t%i\t%i\n" % tuple([i] + list(data[i])))
    #for
#write_table

def annotate(alleles_handle, reference, report_handle, minimum):
    """
    Convert a csv files containing alleles and counts to HGVS descriptions of
    alleles, single variants and variant types.

    :arg alleles_handle: Open handle to the alleles file.
    :type alleles_handle: stream
    :arg reference: The reference sequence.
    :type reference: str
    :arg report_handle: Open writeable handle to the report file.
    :type report_handle: stream
    :arg minimum: Minimum count.
    :type minimum: int
    """
    alleles = collections.defaultdict(lambda: numpy.array([0, 0, 0]))
    raw_vars = collections.defaultdict(lambda: numpy.array([0, 0, 0]))
    classification = collections.defaultdict(lambda: numpy.array([0, 0, 0]))
    service = client.Client(URL).service

    data = map(lambda x: x.strip('\n').split('\t'),
        alleles_handle.readlines()[1:])
    for i in data:
        allele_description = service.descriptionExtract(reference=reference,
            observed=i[0])
        encountered = map(int, (i[1:]))

        alleles[allele_description.description] += encountered
        for j in allele_description.allele:
            for k in j[1]:
                raw_vars[k.hgvs] += encountered
                classification[k.type] += encountered
            #for
    #for

    write_table(alleles, "allele", report_handle, minimum)
    report_handle.write("\n")
    write_table(raw_vars, "variant", report_handle, minimum)
    report_handle.write("\n")
    write_table(classification, "class", report_handle, minimum)
#annotate

def main():
    """
    Main entry point.
    """
    usage = __doc__.split("\n\n\n")
    parser = argparse.ArgumentParser(description=usage[0], epilog=usage[1],
        formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument("alleles", metavar="alleles",
        type=argparse.FileType("r"), help="the alleles file")
    parser.add_argument("reference", metavar="reference", type=str,
        help="sequence of the reference allele")
    parser.add_argument("-r", dest="report", type=argparse.FileType("w"),
        default=sys.stdout, help="name of the report file")
    parser.add_argument("-a", dest="minimum", type=int, default=0,
        help="minimum count (default=%(default)s)")

    args = parser.parse_args()

    try:
        annotate(args.alleles, args.reference, args.report, args.minimum)
    except OSError, error:
        parser.error(error)
#main

if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""
TSSV-Lite: Targeted characterisation of short structural variation.


The library file consists of three tab-separated columns:
- name of the marker pair
- marker 1
- marker 2

Without the -d option, TSSV-Lite will write a table containing a count
of each unique sequence detected to standard output, and a table of
per-marker statistics to the standard error stream. This output can be
redirected to a file using the -o and -r options, respectively.

If the -d option is used, a folder will be created instead, and the
output is written to two files named sequences.csv and statistics.csv in
this folder. Additionally, a subfolder is created for each marker
containing separate tables of the detected sequences and split FASTA
files.

Copyright (c) 2016 Leiden University Medical Center <humgen@lumc.nl>
Copyright (c) 2016 Jeroen F.J. Laros <j.f.j.laros@lumc.nl>
Copyright (c) 2016 Jerry Hoogenboom <j.hoogenboom@nfi.minvenj.nl>
Copyright (c) 2012 Jaap W.F. van der Heijden

Licensed under the MIT license, see the LICENSE file.
"""

import argparse
import math
import sys
import os

from Bio import Seq, SeqIO
from fastools import fastools

from . import ProtectedFileType, version
from .align_pair import align_pair

usage = __doc__.split("\n\n\n")

def parse_library(library_handle, threshold):
    """
    Read a library of flanking sequences from an open file handle.

    :arg library_handle: Open readable handle to a library file.
    :type library_handle: stream
    :arg threshold: Number of allowed mismatches per nucleotide.
    :type threshold: float

    :returns: Dict containing library data.
    :rtype: dict{str: tuple(tuple(str, str), tuple(int, int))}
    """
    return {data[0]: (
        (data[1].upper(), Seq.reverse_complement(data[2].upper())), (
            int(math.ceil(len(data[1]) * threshold)),
            int(math.ceil(len(data[2]) * threshold))))
                for data in (
                    line.split() for line in library_handle.readlines())}

def prepare_output_dir(dir, markers, file_format):
    """
    Create output directories and return a nested dict of output files.

    :arg dir: Output directory name.
    :type dir: str
    :arg markers: Iterable providing the marker names.
    :type markers: iterable
    :arg file_format: Output file format ('fasta' or 'fastq').
    :type file_format: str

    :returns: Nested dict with open writable streams to output files.
    :rtype: dict
    """
    # Create output directories.
    os.mkdir(dir)
    for marker in markers:
        os.mkdir(os.path.join(dir, marker))

    # Open output files.
    return {
        "sequences": open(os.path.join(dir, "sequences.csv"), "w"),
        "statistics": open(os.path.join(dir, "statistics.csv"), "w"),
        "unknown": open(os.path.join(dir, "unknown.f" + file_format[-1]), "w"),
        "markers": {
            marker: {
                "sequences": open(os.path.join(dir, marker,
                    "sequences.csv"), "w"),
                "paired": open(os.path.join(dir, marker,
                    "paired.f" + file_format[-1]), "w"),
                "noend": open(os.path.join(dir, marker,
                    "noend.f" + file_format[-1]), "w"),
                "nostart": open(os.path.join(dir, marker,
                    "nostart.f" + file_format[-1]), "w"),
            } for marker in markers
        }
    }


def find_pair(seq, pair, thresholds, indel_score=1):
    """
    Align a pair of markers to the forward reference sequence.  The
    reverse complement is used to align the second element of the pair
    (which is also reverse complemented).

    :arg seq: Sequence to align to, and its reverse complement.
    :type seq: list[str, str]
    :arg pair: A pair (forward, reverse) of markers to align.
    :type pair: list[str, str]
    :arg thresholds: Maximum number of allowed mismatches.
    :type thresholds: list[int, int]
    :arg indel_score: Penalty score for insertions and deletions per
    nucleotide (optional, default: 1)
    :type indel_score: int

    :returns: Bools indicating whether the two flanks are found, and the
    sequence between the two hits (or None if there was no paired hit).
    :rtype: tuple(list[bool, bool], str)
    """
    seq_up = map(str.upper, seq)
    alignments = align_pair(seq_up[0], seq_up[1], pair, indel_score)
    matches = [False, False]

    if alignments[0][0] <= thresholds[0]:
        # Left marker was found.
        cutout = seq[0][max(0, alignments[0][1]-len(pair[0])):alignments[0][1]]
        if cutout.lower() != cutout:
            matches[0] = True
    if alignments[1][0] <= thresholds[1]:
        # Right marker was found.
        cutout = seq[0][alignments[1][1] : alignments[1][1]+len(pair[1])]
        if cutout.lower() != cutout:
            matches[1] = True

    if matches[0] and matches[1] and alignments[0][1] < alignments[1][1]:
        # Matched pair.
        return matches, seq[0][alignments[0][1] : alignments[1][1]]
    return matches, None


def process_file(input_handle, file_format, library, outfiles=None,
                 indel_score=1):
    """
    Process a FASTA or FASTQ file, attempt to link every read to a
    marker with the given library of flanking sequences.

    If given, the outfiles argument should be a nested dict with the
    following keys, each containing an open writable handle to a FASTA
    or FASTQ file:
    outfiles['unknown'],
    outfiles['markers'][<marker name>]['paired'],
    outfiles['markers'][<marker name>]['noend'],
    outfiles['markers'][<marker name>]['nostart']

    :arg input_handle: Open readable stream with FASTA or FASTQ data.
    :type input_handle: stream
    :arg file_format: The name of the file format ('fasta' or 'fastq').
    :type file_format: str
    :arg library: Dict containing library data.
    :type library: dict{str: tuple(tuple(str, str), tuple(int, int))}
    :arg outfiles: Dict with open writable streams to which FASTA or
    FASTQ data will be written (optional).
    :type outfiles: dict
    :arg indel_score: Penalty score for insertions and deletions per
    nucleotide (optional, default: 1)
    :type indel_score: int

    :returns: The total number of reads, the number of unrecognised
    reads, a dict with counts of flanking sequence hits per marker, and
    a dict with sequences per marker.
    :rtype: tuple(int, int, dict{str: dict{str: int}},
    dict{str: dict{str: list[int, int]}})
    """
    sequences = {marker: {} for marker in library}
    counters = {marker: {key: 0 for key in
            ("fPaired", "rPaired", "fLeft", "rLeft", "fRight", "rRight")}
        for marker in library}
    total_reads = 0
    unrecognised = 0
    for record in SeqIO.parse(input_handle, file_format):
        ref = (str(record.seq), Seq.reverse_complement(str(record.seq)))
        total_reads += 1
        recognised = False

        for marker in library:

            # Search in the forward strand.
            matches, seq = find_pair(ref, library[marker][0],
                library[marker][1], indel_score)
            if matches[0]:
                recognised = True
                counters[marker]["fLeft"] += 1
            if matches[1]:
                recognised = True
                counters[marker]["fRight"] += 1
            if seq is not None:
                counters[marker]["fPaired"] += 1
                if seq not in sequences[marker]:
                    sequences[marker][seq] = [1, 0]
                else:
                    sequences[marker][seq][0] += 1
                if outfiles:
                    SeqIO.write([record],
                        outfiles["markers"][marker]["paired"], file_format)
            elif outfiles:
                if matches[0]:
                    SeqIO.write([record],
                        outfiles["markers"][marker]["noend"], file_format)
                if matches[1]:
                    SeqIO.write([record],
                        outfiles["markers"][marker]["nostart"], file_format)

            # Search in the reverse strand.
            matches, seq = find_pair(ref[-1::-1], library[marker][0],
                library[marker][1], indel_score)
            if matches[0]:
                recognised = True
                counters[marker]["rLeft"] += 1
            if matches[1]:
                recognised = True
                counters[marker]["rRight"] += 1
            if seq is not None:
                counters[marker]["rPaired"] += 1
                if seq not in sequences[marker]:
                    sequences[marker][seq] = [0, 1]
                else:
                    sequences[marker][seq][1] += 1
                if outfiles:
                    SeqIO.write([record],
                        outfiles["markers"][marker]["paired"], file_format)
            elif outfiles:
                if matches[0]:
                    SeqIO.write([record],
                        outfiles["markers"][marker]["noend"], file_format)
                if matches[1]:
                    SeqIO.write([record],
                        outfiles["markers"][marker]["nostart"], file_format)

        if not recognised:
            unrecognised += 1
            if outfiles:
                SeqIO.write([record], outfiles["unknown"], file_format)

    # Count number of unique sequences per marker.
    for marker in library:
        counters[marker]["unique_seqs"] = len(sequences[marker])

    # Return counters and sequences.
    return total_reads, unrecognised, counters, sequences


def make_sequence_tables(sequences, minimum=1):
    """
    Return a dict of tables of the counts of all sequences for all
    markers, with a specified minimum number of reads per sequence.

    :arg sequences: Nested dict of read counts for all sequences of all
    markers.
    :type sequences: dict{str: dict{str: list[int, int]}}
    :arg minimum: The minimum number of reads per sequence to include in
    the table.
    :type minimum: int

    :returns: A tuple with two elements: a tuple of column names and a
    dict with a table of sequences per marker.
    :rtype: tuple(tuple(str), dict{str: list[list]})
    """
    header = ("marker", "sequence", "total", "forward", "reverse")
    tables = {}
    for marker in sequences:
        tables[marker] = []
        for total, sequence in sorted(
                ((sum(sequences[marker][s]), s) for s in sequences[marker]),
                reverse=True):
            if total < minimum:
                break
            tables[marker].append(
                [marker, sequence, total] + sequences[marker][sequence])
    return header, tables


def make_sequence_table(sequences, minimum=1, outfiles=None):
    """
    Return a tab-separated table of the counts of all sequences for all
    markers, with a specified minimum number of reads per sequence.

    :arg sequences: Nested dict of read counts for all sequences of all
    markers.
    :type sequences: dict{str: dict{str: list[int, int]}}
    :arg minimum: The minimum number of reads per sequence to include in
    the table.
    :type minimum: int
    :arg outfiles: Dict with the marker names as keys and open writable
    streams as values, for separate tables for each marker (optional).
    :type outfiles: dict{str: stream}

    :returns: A tab-separated table of the input data.
    :rtype: str
    """
    header, tables = make_sequence_tables(sequences, minimum)
    header = "\t".join(header)
    for marker in sorted(tables):
        tables[marker] = "\n".join(
            "\t".join(map(str, line)) for line in tables[marker])
        if outfiles:
            outfiles[marker].write(
                "\n".join((header, tables[marker])))
    return "\n".join([header] + [tables[marker] for marker in sorted(tables)])


def make_statistics_table(counters):
    return "\n".join([
        "\t".join(("marker", "unique_seqs", "tPaired", "fPaired", "rPaired",
            "tLeft", "fLeft", "rLeft", "tRight", "fRight", "rRight"))] + [
        "\t".join(map(str, [
            marker,
            counters[marker]["unique_seqs"],
            counters[marker]["fPaired"] + counters[marker]["rPaired"],
            counters[marker]["fPaired"],
            counters[marker]["rPaired"],
            counters[marker]["fLeft"] + counters[marker]["rLeft"],
            counters[marker]["fLeft"],
            counters[marker]["rLeft"],
            counters[marker]["fRight"] + counters[marker]["rRight"],
            counters[marker]["fRight"],
            counters[marker]["rRight"]])) for marker in sorted(counters)])


def tssv_lite(input_handle, file_format, library, minimum, dir, tee,
              output_handle, report_handle, indel_score):
    """
    """
    if dir:
        outfiles = prepare_output_dir(dir, library, file_format)
    else:
        outfiles = None

    # Process the input file and obtain counters and sequences.
    total_reads, unrecognised, counters, sequences = process_file(
        input_handle, file_format, library, outfiles, indel_score)

    # Make table of sequences.
    table_of_sequences = make_sequence_table(sequences, minimum,
        None if outfiles is None else
        {m: outfiles["markers"][m]["sequences"] for m in outfiles["markers"]})

    # Make table of statistics.
    statistics = "\n".join((
        make_statistics_table(counters),
        "",  # Empty line.
        "total reads\t%i" % total_reads,
        "unrecognised reads\t%i" % unrecognised))

    if not outfiles or tee or output_handle != sys.stdout:
        output_handle.write(table_of_sequences + "\n")
    if not outfiles or tee or report_handle != sys.stderr:
        report_handle.write(statistics + "\n")
    if outfiles:
        outfiles["sequences"].write(table_of_sequences)
        outfiles["statistics"].write(statistics)


def main():
    """
    Main entry point.
    """
    parser = argparse.ArgumentParser(description=usage[0], epilog=usage[1],
        formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('input_handle', metavar='INPUT',
        type=argparse.FileType('r'), help='input FASTA/FASTQ file')
    parser.add_argument('library_handle', metavar='LIBRARY',
        type=argparse.FileType('r'), help='library of flanking sequences')
    parser.add_argument('-f', '--format', metavar='FORMAT',
        choices=('auto', 'fasta', 'fastq'), default='auto',
        help='file type of the input data: one of %(choices)s '
             '(default=%(default)s)')
    parser.add_argument('-m', '--mismatches', type=float, default=0.08,
        help='mismatches per nucleotide (default=%(default)s)')
    parser.add_argument('-n', '--indel-score', type=int, default=1,
        help='insertions and deletions are penalised this number of times '
             'more heavily than mismatches (default=%(default)s)')
    parser.add_argument('-d', '--dir', type=str,
        help='output directory for verbose output')
    parser.add_argument('-t', '--tee', action='store_true',
        help='if specified together with -d/--dir, output on stdout/stderr '
             'is not suppressed')
    parser.add_argument('-a', '--minimum', type=int, default=0,
        help='minimum count per sequence (default=%(default)s)')
    parser.add_argument('-o', '--output', type=ProtectedFileType('w'),
        default=sys.stdout,
        help='write output to this file (default: write to standard output)')
    parser.add_argument('-r', '--report', type=ProtectedFileType('w'),
        default=sys.stderr,
        help='write a statistical report to this file (default: write to '
             'standard error)')
    parser.add_argument('-v', '--version', action='version',
        version=version(parser.prog))

    args = parser.parse_args()

    try:
        tssv_lite(
            args.input_handle,
            fastools.guess_file_format(args.input_handle)
                if args.format == 'auto' else args.format,
            parse_library(args.library_handle, args.mismatches),
            args.minimum,
            args.dir,
            args.tee,
            args.output,
            args.report,
            args.indel_score)
    except ValueError, error:
        parser.error(error)


if __name__ == '__main__':
    main()



# Patch for swig.
from distutils.command.build import build as _build
_build.sub_commands = [
    ('build_ext', _build.has_ext_modules),
    ('build_py', _build.has_pure_modules),
    ('build_clib', _build.has_c_libraries),
    ('build_scripts', _build.has_scripts)
]

import sys
from setuptools import setup
from distutils.core import Extension

if sys.version_info < (2, 6):
    raise Exception('TSSV requires Python 2.6 or higher.')

# Todo: How does this play with pip freeze requirement files?
requires = ['biopython', 'fastools', 'suds']

# Python 2.6 does not include the argparse module.
try:
    import argparse
except ImportError:
    requires.append('argparse')

import tssv as distmeta

setup(
    name='tssv',
    ext_modules=[Extension('tssv/_sg_align', ['tssv/sg_align.c',
        'tssv/sg_align.i'], swig_opts=[],
        extra_compile_args=['-O3'])],
    py_modules=['tssv.sg_align'],
    version=distmeta.__version__,
    description='Targeted characterisation of short structural variation.',
    long_description=distmeta.__doc__,
    author=distmeta.__author__,
    author_email=distmeta.__contact__,
    url=distmeta.__homepage__,
    license='MIT License',
    platforms=['any'],
    packages=['tssv'],
    install_requires=requires,
    entry_points = {
        'console_scripts': [
            'tssv = tssv.tssv:main',
            'tssvl = tssv.tssv_lite:main',
            'tannotate = tssv.annotate:main',
        ]
    },
    classifiers = [
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering',
    ],
    keywords='bioinformatics'
)
